import os
import shutil

import tableau_utilities.tableau_file.tableau_file_objects as tfo
from tableau_utilities.general.config_column_persona import personas, get_persona_by_attribs, \
    get_persona_by_metadata_local_type
from tableau_utilities.general.cli_styling import Color
from tableau_utilities.general.cli_styling import Symbol
from tableau_utilities.tableau_file.tableau_file import Datasource
from tableau_utilities.tableau_server.tableau_server import TableauServer


def datasource(args, server=None):
    """ Updates a Tableau Datasource locally

    Args:
        args: An argparse args object
        server (TableauServer): A TableauServer object; optional
    """

    # Set variables from args
    # General Args
    debugging_logs = args.debugging_logs
    include_extract = args.include_extract
    save_tds = args.save_tds

    # Datasource Args
    datasource_path = args.file_path
    datasource_id = args.id
    datasource_name = args.name
    project_name = args.project_name
    location = args.location
    enforce_connection = args.enforce_connection

    # Folder/Fields Args
    persona = args.persona
    delete = args.delete
    column_name = args.column_name
    folder_name = args.folder_name
    caption = args.caption
    desc = args.desc
    calculation = args.calculation
    remote_name = args.remote_name
    list_objects = args.list.title() if args.list else None

    # Datasource Connection Args
    conn_type = args.conn_type
    conn_host = args.conn_host
    conn_user = args.conn_user
    conn_role = args.conn_role
    conn_db = args.conn_db
    conn_schema = args.conn_schema
    conn_warehouse = args.conn_warehouse

    # Print Styling
    color = Color()
    symbol = Symbol()

    # Downloads the datasource from Tableau Server if the datasource is not local
    if location == 'online':
        print(f'{color.fg_cyan}...Downloading {datasource_name}...{color.reset}')
        d = server.get_datasource(datasource_id, datasource_name, project_name)
        datasource_path = server.download_datasource(d.id, include_extract=include_extract)
        print(f'{color.fg_green}{symbol.success}  Downloaded Datasource:', f'{color.fg_yellow}{datasource_path}{color.reset}', '\n')

    datasource_file_name = os.path.basename(datasource_path)
    ds = Datasource(datasource_path)

    if save_tds:
        print(f'{color.fg_cyan}...Extracting {datasource_file_name}...{color.reset}')
        save_folder = f'{datasource_file_name} - BEFORE'
        os.makedirs(save_folder, exist_ok=True)
        if ds.extension == 'tds':
            xml_path = os.path.join(save_folder, datasource_file_name)
            shutil.copy(datasource_path, xml_path)
        else:
            xml_path = ds.unzip(extract_to=save_folder, unzip_all=True)
        if debugging_logs:
            print(f'{color.fg_green}{symbol.success}  BEFORE - TDS SAVED TO: {color.fg_yellow}{xml_path}{color.reset}')

    # List each of the objects specified to list
    if list_objects:
        print(f'{color.fg_cyan}{list_objects}:{color.reset}')
    if list_objects == 'Columns':
        for c in ds.columns:
            print(f'  {symbol.arrow_r} '
                  f'{color.fg_yellow}caption:{color.reset} {c.caption} '
                  f'{color.fg_yellow}local-name:{color.reset} {c.name} '
                  f'{color.fg_yellow}persona:{color.reset} {get_persona_by_attribs(c.role, c.type, c.datatype)}')
    if list_objects == 'Folders':
        for f in ds.folders_common.folder:
            print(f'  {symbol.arrow_r} {color.fg_yellow}{f.name}{color.reset}')
    if list_objects == 'Metadata':
        for m in ds.connection.metadata_records:
            print(f'  {symbol.arrow_r} '
                  f'{color.fg_yellow}local-name:{color.reset} {m.local_name} '
                  f'{color.fg_yellow}remote-name:{color.reset} {m.remote_name} '
                  f'{color.fg_yellow}persona:{color.reset} {get_persona_by_metadata_local_type(m.local_type)}')
    if list_objects == 'Connections':
        for c in ds.connection.named_connections:
            print(f'  {symbol.arrow_r} {c.connection.dict()}')

    # Add / modify a specified column
    if column_name and not delete:
        # Column name needs to be enclosed in brackets
        if debugging_logs:
            print('Going to add/update column:', column_name)
        column = ds.columns.get(column_name)
        if persona:
            persona = personas.get(persona.lower(), {})
        else:
            persona = dict()

        if not column:
            if not persona:
                raise Exception('Column does not exist, and more args are need to add a new column.\n'
                                f'Minimum required args: {color.fg_yellow}--column_name --persona{color.reset}')
            column = tfo.Column(
                name=column_name,
                role=persona['role'],
                datatype=persona['datatype'],
                type=persona['role_type'],
            )
            print(f'{color.fg_cyan}Creating new column for {column_name}:{color.reset} {column.dict()}')
        else:
            print(f'{color.fg_cyan}Updating existing column:{color.reset}\n  {column.dict()}')

        column.caption = caption or column.caption
        column.role = persona.get('role') or column.role
        column.type = persona.get('role_type') or column.type
        column.datatype = persona.get('datatype') or column.datatype
        column.desc = desc or column.desc
        column.calculation = calculation or column.calculation
        ds.enforce_column(column, remote_name=remote_name, folder_name=folder_name)

    # Add a folder if it was specified and does not exist already
    if folder_name and not ds.folders_common.get(folder_name) and not delete:
        if debugging_logs:
            print(f'Going to add folder: {color.fg_cyan}{folder_name}{color.reset}')
        ds.folders_common.add(tfo.Folder(name=folder_name))

    # Delete specified object
    if delete == 'column':
        ds.columns.delete(column_name)
    if delete == 'folder':
        ds.folders_common.folder.delete(folder_name)

    # Enforce Connection
    if enforce_connection:
        if debugging_logs:
            print(f'Updating the datasource connection: {color.fg_cyan}{conn_type}{color.reset}')
        connection = ds.connection.get(conn_type)
        if not connection and debugging_logs:
            print(f'Datasource does not contain a connection of type: {conn_type}')
        else:
            connection.class_name = conn_type or connection.class_name
            connection.server = conn_host or connection.server
            connection.username = conn_user or connection.username
            connection.service = conn_role or connection.service
            connection.dbname = conn_db or connection.dbname
            connection.schema = conn_schema or connection.schema
            connection.warehouse = conn_warehouse or connection.warehouse
            ds.connection.update(connection)

    # Save the datasource if an edit may have happened
    if column_name or folder_name or delete or enforce_connection:
        ds.save()
        print(f'{color.fg_green}{symbol.success}  Saved changes to: {color.fg_yellow}{datasource_path}{color.reset}')

    if save_tds:
        print(f'{color.fg_cyan}...Extracting {datasource_file_name}...{color.reset}')
        save_folder = f'{datasource_file_name} - AFTER'
        os.makedirs(save_folder, exist_ok=True)
        if ds.extension == 'tds':
            xml_path = os.path.join(save_folder, datasource_file_name)
            shutil.copy(datasource_path, xml_path)
        else:
            xml_path = ds.unzip(extract_to=save_folder, unzip_all=True)
        if debugging_logs:
            print(f'{color.fg_green}{symbol.success}  AFTER - TDS SAVED TO: {color.fg_yellow}{xml_path}{color.reset}')



import json
from tableau_utilities.general.cli_styling import Color, Symbol
from tableau_utilities.scripts.gen_config import load_csv_with_definitions, generate_config


def read_file(file_path):
    """ Read a JSON file to a dictionary

    Args:
        file_path (str): The path of the file to read

    """
    with open(file_path, "r") as infile:
        config: dict = json.load(infile)

    return config


def write_file(file_name, config, debugging_logs=False):
    """ Write a dictionary to a JSON file

    Args:
        file_name: The name of the file to write
        config: The dictionary to write to the file

    """
    with open(file_name, "w") as outfile:
        json.dump(config, outfile)

    if debugging_logs:
        print('CONFIG PATH:', file_name)


def add_definitions_mapping(config, definitions_mapping):
    """ Adds definitions from a mapping to the config. Chooses the definition from the mapping if needed

    Args:
        config: A datasource config
        definitions_mapping: A mapping of columns to definitions

    """
    for column, definition in definitions_mapping.items():
        if len(definition) > 0 and column in config:
            config[column]['description'] = definition
    return config


def merge_2_configs(existing_config, additional_config, debugging_logs=False):
    """ Takes 2 configs and adds information from the additional_cong to the existing_config
    The output of the merged config should  be merged into the existing config in a PR
    This assumes that the user is merging the same type of config

    Args:
        existing_config (dict): The current existing config
        additional_config (dict): The additional config to add
        debugging_logs (bool): True to print debugging logs

    """
    for column_name, column_details in additional_config.items():
        # If the column doesn't exist add it
        if column_name not in existing_config:
            if debugging_logs:
                print("ADDING COLUMN", column_name, column_details)
            existing_config[column_name] = column_details
        else:
            if debugging_logs:
                print('ALTERING COLUMN:', column_name)
                print('\tEXISTING DETAIL:', existing_config[column_name])
                print('\tADDITIONAL DETAIL:', column_details)

            # Replace existing config attributes
            for attribute_name, attribute_value in column_details.items():
                if str(attribute_value).strip() not in ['', 'None'] and attribute_name != 'datasources':
                    if debugging_logs:
                        print(
                            '\tUPDATING ATTRIBUTE', attribute_name,
                            'FROM:', existing_config[column_name][attribute_name],
                            '-> TO:', attribute_value
                        )
                    existing_config[column_name][attribute_name] = attribute_value
            # Remove the existing column datasource if it is included from the additional config
            existing_config[column_name]['datasources'] = [
                d for d in existing_config[column_name]['datasources']
                if d['name'] not in [d['name'] for d in column_details['datasources']]
            ]
            # Add all datasources from the additional column to existing column datasources
            for datasource in column_details['datasources']:
                if debugging_logs:
                    print('\tADDING DATASOURCE:', datasource)
                existing_config[column_name]['datasources'].append(datasource)

            if debugging_logs:
                print('\tCOLUMN ATTRIBUTES SET TO:', existing_config[column_name])

    return existing_config


def sort_config(config, debugging_logs):
    """ Takes a config and returns it sorted by key and with the datasources list sorted

    Args:
        config (dict): A datasource config
        debugging_logs: Prints info to console if true

    Returns:
        config (dict): With the list of datasources sorted.  With the config sorted by column_name (top key)

    """

    # Sort the list of datasources alphabetically
    for k, v in config.items():

        if debugging_logs:
            print('KEY', k)
            print('CONGIG', v)
            print('DATASOURCES', v['datasources'])

        sorted_datasources = sorted(v['datasources'], key=lambda d: d['name'])

        if debugging_logs:
            print('DATASOURCS SORTED', sorted_datasources)

        config[k]['datasources'] = sorted_datasources

    # Sort the top level keys alphabetically
    config = dict(sorted(config.items()))

    return config


def read_merge_write(existing_config_path, additional_config_path, output_config_path, debugging_logs):
    # Print Styling
    color = Color()
    symbol = Symbol()

    # Read files
    existing_config = read_file(existing_config_path)
    additional_config = read_file(additional_config_path)

    # Merge
    new_config = merge_2_configs(existing_config, additional_config, debugging_logs)

    # Sort and write the merged config
    new_config = sort_config(new_config, debugging_logs)
    write_file(file_name=output_config_path, config=new_config, debugging_logs=debugging_logs)

    print(f'{color.fg_yellow}EXISTING CONFIG {symbol.arrow_r} {color.fg_grey}{existing_config_path}{color.reset}')
    print(f'{color.fg_yellow}ADDITIONAL CONFIG {symbol.arrow_r} {color.fg_grey}{additional_config_path}{color.reset}')
    print(f'{color.fg_green}{symbol.success}  MERGED CONFIG {symbol.arrow_r} '
          f'{color.fg_grey}{output_config_path}{color.reset}')


def merge_configs(args, server=None):
    """ Merges 2 configs for datasource automation

    """

    # Set variables from args
    existing_config_path = args.existing_config
    additional_config_path = args.additional_config
    definitions_csv_path = args.definitions_csv
    merge_with = args.merge_with
    file_name = f'{args.merged_config}.json'
    target_directory = args.target_directory
    debugging_logs = args.debugging_logs
    existing_config = args.existing_config

    # Print Styling
    color = Color()
    symbol = Symbol()

    # Merge 2 configs
    if merge_with == 'config':
        read_merge_write(existing_config_path, additional_config_path, file_name, debugging_logs)

    # Merge a config with a definitions csv
    elif merge_with == 'csv':
        # Read files
        existing_config = read_file(existing_config)
        definitions_mapping = load_csv_with_definitions(file=definitions_csv_path)
        # Merge
        new_config = add_definitions_mapping(existing_config, definitions_mapping)
        # Sort and write the merged config
        new_config = sort_config(new_config, debugging_logs)
        write_file(file_name=file_name, config=new_config, debugging_logs=debugging_logs)

        print(f'{color.fg_yellow}EXISTING CONFIG {symbol.arrow_r} '
              f'{color.fg_grey}{existing_config_path}{color.reset}')
        print(f'{color.fg_yellow}ADDITIONAL CONFIG {symbol.arrow_r} '
              f'{color.fg_grey}{additional_config_path}{color.reset}')
        print(f'{color.fg_green}{symbol.success}  MERGED CONFIG {symbol.arrow_r} '
              f'{color.fg_grey}{file_name}{color.reset}')

    elif merge_with == 'generate_merge_all':
        # Generate the configs and return the paths of where they are
        new_column_config_path, new_calculated_column_config_path = generate_config(args, server)
        print(f'{color.fg_green}{symbol.success}  GENERATED CONFIGS {symbol.arrow_r} '
              f'{color.fg_grey}{new_column_config_path} {symbol.sep} {new_calculated_column_config_path}{color.reset}')
        print(f'{color.fg_yellow}TARGET DIRECTORY {symbol.arrow_r} {color.fg_grey}{target_directory}{color.reset}')

        existing_column_config_path = f'{target_directory}column_config.json'
        existing_calc_config_path = f'{target_directory}tableau_calc_config.json'

        read_merge_write(existing_config_path=existing_column_config_path,
                         additional_config_path=new_column_config_path,
                         output_config_path=existing_column_config_path,
                         debugging_logs=debugging_logs)
        read_merge_write(existing_config_path=existing_calc_config_path,
                         additional_config_path=new_calculated_column_config_path,
                         output_config_path=existing_calc_config_path,
                         debugging_logs=debugging_logs)







import xml.etree.ElementTree as ET
import os
import shutil

import xmltodict
from zipfile import ZipFile

import tableau_utilities.tableau_file.tableau_file_objects as tfo
from tableau_utilities.general.funcs import transform_tableau_object


class TableauFileError(Exception):
    """ A minimum viable exception. """

    def __init__(self, message):
        self.message = message


class TableauFile:
    """ The base class for a Tableau file, i.e. Datasource or Workbook. """

    def __init__(self, file_path):
        """
        Args:
            file_path (str): Path to a Tableau file

        """
        self.file_path = os.path.abspath(file_path)
        self.file_directory = os.path.dirname(self.file_path)
        self.file_basename = os.path.basename(self.file_path)
        self.extension = file_path.split('.')[-1]
        self.file_name = self.file_basename.replace(f'.{self.extension}', '')
        ''' Set on init '''
        self._tree: ET.ElementTree
        self._root: ET.Element
        self.has_extract_data: bool = False
        self.__extract_xml()

    def __extract_xml(self, path=None):
        """ Extracts the XML from a Tableau file.

        Args:
            path (str): The path to the zipped Tableau file

        Returns: The contents of the Tableau File
        """
        if not path:
            path = self.file_path

        if self.extension in ['tdsx', 'twbx']:
            with ZipFile(path) as zip_file:
                for z in zip_file.filelist:
                    if z.filename.split('.')[-1] not in ['tds', 'twb']:
                        self.has_extract_data = True
                        continue
                    self._tree = ET.parse(zip_file.open(z.filename))
                    self._root = self._tree.getroot()
        else:
            self._tree = ET.parse(path)
            self._root = self._tree.getroot()

    def unzip(self, unzip_all=False, extract_to=None):
        """ Unzips the Tableau File.

        Args:
            unzip_all (bool): True to unzip all zipped files
            extract_to: Override the source file directory and save the file to another location

        Returns: The path to the unzipped Tableau File
        """

        if extract_to is not None:
            file_dir = extract_to
        else:
            file_dir = self.file_directory

        tableau_file_path = None
        with ZipFile(self.file_path) as zip_file:
            for z in zip_file.filelist:
                ext = z.filename.split('.')[-1]
                if unzip_all:
                    zip_file.extract(member=z, path=file_dir)
                    if ext in ['tds', 'twb']:
                        tableau_file_path = os.path.join(file_dir, z.filename)
                elif not unzip_all and ext in ['tds', 'twb']:
                    zip_file.extract(member=z, path=file_dir)
                    tableau_file_path = os.path.join(file_dir, z.filename)
        return tableau_file_path

    def save(self):
        """ Save/Update the Tableau file with the XML changes made """
        if self.extension in ['tdsx', 'twbx']:
            # Rebuild the TDSX / TWBX archive file, with the updated archived TDS / TWB
            # Move the file into a temporary folder while updating
            temp_folder = os.path.join(self.file_directory, f'__TEMP_{self.file_name}')
            os.makedirs(temp_folder, exist_ok=False)
            temp_path = os.path.join(temp_folder, self.file_basename)
            shutil.move(self.file_path, temp_path)
            # Unzip the zipped files
            extracted_files = list()
            with ZipFile(temp_path) as z:
                for f in z.filelist:
                    ext = f.filename.split('.')[-1]
                    path = z.extract(member=f, path=temp_folder)
                    extracted_files.append(path)
                    if ext in ['tds', 'twb']:
                        xml_path = path
            # Update XML file
            self._tree.write(xml_path, encoding="utf-8", xml_declaration=True)
            # Repack the unzipped file
            with ZipFile(temp_path, 'w') as z:
                for file in extracted_files:
                    arcname = file.split(temp_folder)[-1]
                    z.write(file, arcname=arcname)
            # Move file back to the original folder and remove any unpacked contents
            shutil.move(temp_path, self.file_path)
            shutil.rmtree(temp_folder)
        else:
            # Update the Tableau file's contents
            self._tree.write(self.file_path, encoding="utf-8", xml_declaration=True)


class Datasource(TableauFile):
    """
        A class representation of a Tableau Datasource.
        Used to update a Tableau Datasource by interacting with various elements,
        such as Columns, Folders, Connections, Metadata, etc.
    """

    def __init__(self, file_path):
        """
        Args:
            file_path (str): Path to a Tableau Datasource file; tds or tdsx
        """
        super().__init__(file_path)
        # Validate the file on initialization
        if self.extension not in ['tds', 'tdsx']:
            raise TableauFileError('File must be TDS or TDSX')

        self.connection: tfo.ParentConnection = self.__get_section(tfo.ParentConnection)
        self.aliases: tfo.Aliases = self.__get_section(tfo.Aliases)
        self.columns: tfo.TableauFileObjects[tfo.Column] = self.__get_section(tfo.Column, enforce_list=True)
        self.column_instance: tfo.ColumnInstance = self.__get_section(tfo.ColumnInstance)
        self.drill_paths: tfo.DrillPaths = self.__get_section(tfo.DrillPaths)
        self.folders_common: tfo.FoldersCommon = self.__get_section(tfo.FoldersCommon)
        self.date_options: tfo.DateOptions = self.__get_section(tfo.DateOptions)
        self.extract: tfo.Extract = self.__get_section(tfo.Extract)

    def sections(self):
        """ Yields each section defined in the class, for iteration """
        yield self.connection
        yield self.aliases
        yield self.columns
        yield self.column_instance
        yield self.drill_paths
        yield self.folders_common
        yield self.date_options
        yield self.extract

    def __get_section(self, obj, enforce_list=False):
        """ Sets DatasourceItems for each section

        Args:
            obj (type[tfo.TableauFileObject]): A Tableau File Object; ParentConnection, Column, etc
            enforce_list (bool): True if the section should be a TableauFileObjects list
        """
        parent = self._root.find('.')
        # Gets elements within the parent element, with the appropriate section.tag
        section: list[dict] = list()
        for element in parent:
            if element.tag.endswith(f'true...{obj.tag}') or element.tag == obj.tag:
                item = xmltodict.parse(ET.tostring(element))[element.tag]
                if not item:
                    continue
                new_item = transform_tableau_object(item)
                try:
                    section.append(obj(**new_item))
                except TypeError as err:
                    raise TableauFileError(f'{err}\n\nPre-transform {obj.tag} attributes: {item}') from err
        if len(section) > 1 or len(section) == 1 and enforce_list:
            return tfo.TableauFileObjects(section, item_class=obj, tag=obj.tag)
        if len(section) == 1:
            return section[0]
        if enforce_list:
            return tfo.TableauFileObjects(item_class=obj, tag=obj.tag)
        return obj()

    def enforce_column(self, column, folder_name=None, remote_name=None):
        """
            Enforces a column by:
                - Adding the column if it doesn't exist, otherwise updating it to match the column
                - Adding the column's corresponding folder-item to the appropriate folder, if it doesn't exist
                    - Create the folder if it doesn't exist
                - Updating the metadata local-name to map to the column name
                - Adding the column mapping to the mapping cols, if it doesn't exist

        Args:
            column (tfo.Column): The TableFile Column object
            remote_name (str): The name of the column from the connection (not required for Tableau Calculations),
             i.e. the SQL alias if the connection is a SQL query
            folder_name (str): The name of the folder that the column should be in

        """
        # Add Column
        if column not in self.columns:
            self.columns.add(column)
        # Update the Column
        else:
            self.columns.update(column)

        # Add Folder / FolderItem for the column, if folder_name was provided
        if folder_name:
            # Remove the column's folder-item for preview folder, if it will be moved to a new folder
            current_folder = [f for f in self.folders_common.folder if f.folder_item.get(column.name)]
            if current_folder and current_folder[0].name != folder_name:
                current_folder[0].folder_item.delete(column.name)
                self.folders_common.folder.update(current_folder[0])
            # Add column to the specified folder
            folder = self.folders_common.folder.get(folder_name)
            folder_item = tfo.FolderItem(name=column.name)
            if folder and folder_item not in folder.folder_item:
                folder.folder_item.append(folder_item)
                self.folders_common.folder.update(folder)
            elif not folder:
                self.folders_common.folder.add(tfo.Folder(name=folder_name, folder_item=[folder_item]))

        # If a remote_name was provided, and the column is not a Tableau Calculation - enforce metadata
        if not remote_name or column.calculation:
            return None

        # Update Connection MetadataRecords & MappingCols
        connection_record = self.connection.metadata_records.get(remote_name)
        if not connection_record:
            raise TableauFileError(f'Remote name provided is not in the metadata of the connection: {remote_name}')
        connection_record.local_name = column.name
        self.connection.metadata_records.update(connection_record)
        conn_col = tfo.MappingCol(key=column.name, value=f'{connection_record.parent_name}.[{remote_name}]')
        # Update Connection MappingCols
        if conn_col not in self.connection.cols:
            # Update the Connection MappingCol if key or value is different
            found = False
            for col in self.connection.cols:
                if conn_col.key == col.key or conn_col.value == col.value:
                    found = True
                    self.connection.cols[col].key = conn_col.key
                    self.connection.cols[col].value = conn_col.value
            # Otherwise, Add Connection MappingCol
            if not found:
                self.connection.cols.append(conn_col)

        # Update Extract MetadataRecords & MappingCols
        if not self.extract:
            return None
        extract_record = self.extract.connection.metadata_records.get(remote_name)
        if not extract_record:
            raise TableauFileError(f'Remote name provided is not in the metadata of the extract: {remote_name}')
        extract_record.local_name = column.name
        self.extract.connection.metadata_records.update(extract_record)
        extract_col = tfo.MappingCol(key=column.name, value=f'{extract_record.parent_name}.[{remote_name}]')
        # Update Extract MappingCols
        if extract_col not in self.extract.connection.cols:
            # Update the Extract MappingCol if key or value is different
            found = False
            for col in self.extract.connection.cols:
                if extract_col.key == col.key or extract_col.value == col.value:
                    found = True
                    self.extract.connection.cols[col].key = extract_col.key
                    self.extract.connection.cols[col].value = extract_col.value
            # Otherwise, Add Extract MappingCol
            if not found:
                self.extract.connection.cols.append(extract_col)

    def save(self):
        """ Save all changes made to each section of the Datasource """
        parent = self._root.find('.')
        ending_index = -1
        for section in self.sections():
            if not section:
                continue
            # Find all elements within the parent element, and the index of those elements
            elements = [(idx, element)
                        for idx, element in enumerate(parent)
                        if element.tag.endswith(f'true...{section.tag}') or element.tag == section.tag]
            # If there are no existing element(s), the index will be for the previous ending_index (default == -1)
            starting_index = elements[0][0] if elements else ending_index
            ending_index = elements[-1][0] + 1 if elements else starting_index
            # Remove the existing items
            for _, e in elements:
                parent.remove(e)
            # Insert the new / updated items
            if isinstance(section, tfo.TableauFileObjects):
                section.reverse()
                for idx, item in enumerate(section, 1):
                    parent.insert(starting_index, item.xml())
                    ending_index = starting_index + idx
            else:
                parent.insert(starting_index, section.xml())
        super().save()


if __name__ == '__main__':
    # Params
    ds_path = 'downloads/Users + Orgs.tdsx'

    unzip = False
    unzip_all_files = False

    ds = Datasource(ds_path)
    if unzip:
        ds.unzip(unzip_all=unzip_all_files)

    print(ds.columns.get('[USER_ID]'))



from dataclasses import dataclass, fields


@dataclass
class Color:
    """ Applied standard colors for use when printing to the terminal """
    reset: str = '\033[0m'
    # Foreground colors
    fg_black: str = '\033[30m'
    fg_red: str = '\033[31m'
    fg_green: str = '\033[32m'
    fg_yellow: str = '\033[33m'
    fg_blue: str = '\033[34m'
    fg_magenta: str = '\033[35m'
    fg_cyan: str = '\033[36m'
    fg_white: str = '\033[37m'
    fg_grey: str = '\033[38;5;8m'
    # Background colors
    bg_black: str = '\033[40m'
    bg_red: str = '\033[41m'
    bg_green: str = '\033[42m'
    bg_yellow: str = '\033[43m'
    bg_blue: str = '\033[44m'
    bg_magenta: str = '\033[45m'
    bg_cyan: str = '\033[46m'
    bg_white: str = '\033[47m'
    bg_grey: str = '\033[48;5;8m'

    def __getattr__(self, item):
        expected_fields = [f.name for f in fields(self)]
        if item not in expected_fields:
            raise AttributeError(f'Provided color does not exist: {item}')
        return getattr(self, item)


@dataclass
class Symbol:
    arrow_r: str = '→'
    arrow_l: str = '←'
    line: str = '━'
    success: str = '✅'
    warning: str = '⚠️'
    fail: str = '🚫'
    sep: str = '║'


def color_print(*args, fg=None, bg=None):
    """ Prints arguments to the terminal, in the specified foreground/background color.

    Args:
        fg (str): The foreground color to print
        bg (str): The background color to print
    """
    if not fg and not bg:
        print(*args)
        return None

    c = Color()
    style = ''.join([getattr(c, f'{k}_{v}') for k, v in {'fg': fg, 'bg': bg}.items() if v])
    print(f'{style}{" ".join(args)}{c.reset}')


if __name__ == '__main__':
    c = Color()
    s = Symbol()
    print(s)
    print(f'{s.arrow_r} {c.fg_red}This is fg_red{c.reset}')
    color_print(f'{s.line} This is fg_green', fg='green')
    color_print('This is fg_grey', fg='grey')
    color_print('This is fg_cyan', fg='cyan')
    color_print('This is bg_blue', bg='blue')
    color_print('This is fg_red and bg_blue', fg='red', bg='blue')
    print(f'This {s.sep} is normal')
    # color_print('This is wont print', color='bad_color')
    print(f'{c.bad_color}This is wont print{c.reset}')



personas = {
    "string_dimension": {
        "role": "dimension",
        "role_type": "nominal",
        "datatype": "string"
    },
    "date_dimension": {
        "role": "dimension",
        "role_type": "ordinal",
        "datatype": "date"
    },
    "datetime_dimension": {
        "role": "dimension",
        "role_type": "ordinal",
        "datatype": "datetime"
    },
    "date_measure": {
        "role": "measure",
        "role_type": "ordinal",
        "datatype": "date"
    },
    "datetime_measure": {
        "role": "measure",
        "role_type": "ordinal",
        "datatype": "datetime"
    },
    "discrete_number_dimension": {
        "role": "dimension",
        "role_type": "ordinal",
        "datatype": "integer"
    },
    "continuous_number_dimension": {
        "role": "dimension",
        "role_type": "quantitative",
        "datatype": "integer"
    },
    "discrete_number_measure": {
        "role": "measure",
        "role_type": "ordinal",
        "datatype": "integer"
    },
    "continuous_number_measure": {
        "role": "measure",
        "role_type": "quantitative",
        "datatype": "integer"
    },
    "discrete_decimal_dimension": {
        "role": "dimension",
        "role_type": "ordinal",
        "datatype": "real"
    },
    "continuous_decimal_dimension": {
        "role": "dimension",
        "role_type": "quantitative",
        "datatype": "real"
    },
    "discrete_decimal_measure": {
        "role": "measure",
        "role_type": "ordinal",
        "datatype": "real"
    },
    "continuous_decimal_measure": {
        "role": "measure",
        "role_type": "quantitative",
        "datatype": "real"
    },
    "boolean_dimension": {
        "role": "dimension",
        "role_type": "nominal",
        "datatype": "boolean"
    },
    "boolean_measure": {
        "role": "measure",
        "role_type": "nominal",
        "datatype": "boolean"
    }
}


def get_persona_by_attribs(role, role_type, datatype):
    """ Gets the name of the Persona based on the attributes provided.

    Args:
        role (str): The role of the persona
        role_type (str): The type of the persona
        datatype (str): The datatype of the persona

    Returns: The name of the persona
    """
    search_dict = {'role': role, 'role_type': role_type, 'datatype': datatype}
    for name, attribs in personas.items():
        if attribs == search_dict:
            return name
    return None


def get_persona_by_metadata_local_type(local_type):
    """ Gets the name of the persona based on the local_type of a metadata record.

    Args:
        local_type: The local_type attribute of a metadata record

    Returns: The name of the persona
    """
    # I think it's low risk to assume these data types are all dimensions
    if local_type == 'string':
        return 'string_dimension'
    elif local_type == 'date':
        return 'date_dimension'
    elif local_type == 'datetime':
        return 'datetime_dimension'
    elif local_type == 'boolean':
        return 'boolean_dimension'
    # Makes the assumption that numbers are discrete measures so that a user can't do accidental math on fields
    elif local_type == 'integer':
        return 'discrete_number_dimension'
    elif local_type == 'real':
        return 'discrete_decimal_dimension'

    return None


if __name__ == '__main__':
    print(get_persona_by_attribs('measure', 'nominal', 'boolean'))
    print(get_persona_by_metadata_local_type('string'))



import os
import yaml
import shutil
import xml.etree.ElementTree as ET
from pprint import pprint, isreadable
from tableau_utilities import Datasource, TableauServer
from tableau_utilities import tableau_file_objects as tfo


if __name__ == '__main__':
    with open('settings.yaml') as f:
        settings = yaml.safe_load(f)
    tmp_folder = 'tmp_downloads'
    # Cleanup lingering files, if there are any
    # shutil.rmtree(tmp_folder, ignore_errors=True)

    # Create a temp directory for testing
    os.makedirs(tmp_folder, exist_ok=True)
    os.chdir(tmp_folder)

    #  ### ### Testing here ### ###

    # datasource = Datasource('../resources/test_data_source.tdsx')
    # datasource.unzip()
    # datasource = Datasource('../resources/no_folder.tdsx')
    # datasource.unzip()
    datasource = Datasource('test_data_source.tds')
    # datasource = Datasource('no_folder (local copy).tds')
    # folder = tfo.Folder('Test')
    # datasource.folders_common.add(folder)
    # datasource.save()
    # datasource = Datasource('test_data_source.tds')
    # datasource = Datasource('no_folder (local copy).tds')
    pprint(datasource.folders_common)
    # datasource.save()
    # datasource.unzip()
    # Connect to Tableau Server
    # ts = TableauServer(**settings['tableau_login'])
    # ts.get_datasource(datasource_id='Jobs')
    # datasources = [(d.id, d.name) for d in ts.get_datasources()]
    # pprint(datasources)
    # Loop through the datasources
    # for d in datasources:
    # for path in os.listdir():
    #     if path != 'Jobs.tdsx':
    #         continue
        # path = ts.download_datasource(datasource_id=d.id, include_extract=True)
        # print('Downloaded:', path)
        # datasource = Datasource(path)
        # pprint(datasource.folders_common.folder)
        # print(datasource.columns['job_id'].caption)
        # datasource.columns['job_id'].caption = 'Job ID RENAMED'
        # datasource._tree.write(path.replace('.tdsx',  '.tds'))
        # datasource.save()
        # datasource = Datasource(path)
        # print(datasource.columns['job_id'].caption)
        # column = tfo.Column(
        #     name='FRIENDLY_NAME',
        #     caption='Friendly Name',
        #     datatype='string',
        #     type='ordinal',
        #     role='dimension',
        #     desc='Nice and friendly',
        # )
        # datasource.enforce_column(column, remote_name='ORG_ID', folder_name='Org')
        # print(datasource.folders.get('Org'))
        # print(datasource.datasource_metadata.get('ORG_ID'))
        # print(datasource.extract_metadata.get('ORG_ID'))
        # print(datasource.datasource_mapping_cols.get(column.name))
        # print(datasource.extract_mapping_cols.get(column.name))

    # ### ### ### ### ### ### ###

    # Cleanup lingering files, if there are any
    # os.chdir('..')
    # shutil.rmtree(tmp_folder)



from setuptools import setup

with open('README.md') as f:
    readme = f.read()

setup(
    author="Justin Grilli",
    author_email="justin.grilli@gmail.com",
    license='MIT',
    url='http://pypi.python.org/pypi/tableau-utilities/',
    description='Utility for maintaining Tableau objects',
    long_description=readme,
    long_description_content_type='text/markdown',
    name="tableau_utilities",
    version="2.0.67",
    packages=[
        'tableau_utilities',
        'tableau_utilities.general',
        'tableau_utilities.tableau_file',
        'tableau_utilities.tableau_server',
        'tableau_utilities.scripts'
    ],
    package_data={'tableau_utilities': ['tableau_file/*.yml']},
    include_package_data=True,
    install_requires=['xmltodict>=0.12.0,<1.0.0',
                      'pyyaml>=6.0,<7.0.0',
                      'requests>=2.27.1,<3.0.0',
                      'pandas>=1.4.1,<2.0.0',
                      'tabulate>=0.8.9,<1.0.0'],
    entry_points={
        'console_scripts': [
            'tableau_utilities = tableau_utilities.scripts.cli:main',
        ]
    }
)



from .scripts import cli
from .tableau_file.tableau_file import TableauFileError, Datasource
from .tableau_file import tableau_file_objects
from .tableau_server.tableau_server import TableauConnectionError, TableauServer
from .tableau_server import tableau_server_objects



import xmltodict
from datetime import datetime
from dataclasses import dataclass, asdict, astuple


@dataclass
class ServerObject:
    """
        A Tableau Server object is any object in Tableau Online/Server,
        i.e. Datasources, Workbooks, Users, Groups, Projects, etc.

        This is the base class for a Tableau Server object,
        with the minimum required attributes and functionality.
        Child classes will inherit the functionality provided by this base class.

        The data types of attributes from child class will be converted to the appropriate type,
        if it is provided as a string instead.
    """
    id: str = None

    def __existing_str_attr(self, attr: str):
        """ Returns: True if the attribute exists and is a string """
        return hasattr(self, attr) and isinstance(getattr(self, attr), str)

    def __to_datetime(self, attr: str):
        """
            Set the attribute to a datetime,
            if the class has the attribute,
            and if the attribute is a string.
        """
        if self.__existing_str_attr(attr):
            setattr(self, attr, datetime.strptime(getattr(self, attr), '%Y-%m-%dT%H:%M:%SZ'))

    def __to_int(self, attr: str):
        """
            Set the attribute to an integer,
            if the class has the attribute,
            and if the attribute is a string.
        """
        if self.__existing_str_attr(attr):
            setattr(self, attr, int(getattr(self, attr)))

    def __to_bool(self, attr: str):
        """
            Set the attribute to a boolean,
            if the class has the attribute,
            and if the attribute is a string.
        """
        if self.__existing_str_attr(attr):
            setattr(self, attr, getattr(self, attr).lower() == 'true')

    def __post_init__(self):
        # Convert to Datetime
        self.__to_datetime('created_at')
        self.__to_datetime('completed_at')
        self.__to_datetime('last_login')
        self.__to_datetime('next_run_at')
        self.__to_datetime('updated_at')
        # Convert to Boolean
        self.__to_bool('attach_image')
        self.__to_bool('attach_pdf')
        self.__to_bool('content_send_if_view_empty')
        self.__to_bool('data_acceleration_config_acceleration_enabled')
        self.__to_bool('embed_password')
        self.__to_bool('encrypt_extracts')
        self.__to_bool('has_alert')
        self.__to_bool('has_extracts')
        self.__to_bool('is_certified')
        self.__to_bool('is_published')
        self.__to_bool('is_embedded')
        self.__to_bool('query_tagging_enabled')
        self.__to_bool('show_tabs')
        self.__to_bool('suspended')
        self.__to_bool('top_level_project')
        self.__to_bool('use_remote_query_agent')
        self.__to_bool('writeable')
        # Convert to Integer
        self.__to_int('connected_workbooks_count')
        self.__to_int('contents_counts_datasource_count')
        self.__to_int('contents_counts_project_count')
        self.__to_int('contents_counts_view_count')
        self.__to_int('contents_counts_workbook_count')
        self.__to_int('favorites_total')
        self.__to_int('finish_code')
        self.__to_int('port')
        self.__to_int('priority')
        self.__to_int('server_port')
        self.__to_int('sheet_count')
        self.__to_int('size')
        self.__to_int('usage_total_view_count')
        self.__to_int('user_count')

    def dict(self):
        dictionary = asdict(self)
        keys = list(dictionary.keys())
        for key in keys:
            if dictionary[key] is None:
                del dictionary[key]
        return dictionary


@dataclass
class User(ServerObject):
    """ A TableauServer User object """
    name: str = None
    full_name: str = None
    domain_name: str = None
    auth_setting: str = None
    email: str = None
    email_domain: str = None
    external_auth_user_id: str = None
    site_role: str = None
    locale: str = None
    language: str = None
    last_login: datetime = None

    def __post_init__(self):
        if not self.email_domain and self.email:
            self.email_domain = self.email.lower().split('@')[-1]
        super().__post_init__()

    def __hash__(self):
        return hash(str(astuple(self)))


@dataclass
class Project(ServerObject):
    """ A TableauServer Project object """
    name: str = None
    content_permissions: str = 'LockedToProject'
    description: str = ''
    created_at: datetime = None
    updated_at: datetime = None
    owner_id: str = None
    contents_counts_project_count: int = None
    contents_counts_view_count: int = None
    contents_counts_datasource_count: int = None
    contents_counts_workbook_count: int = None
    parent_project_id: str = None
    controlling_permissions_project_id: str = None
    top_level_project: bool = None
    writeable: bool = None

    def __hash__(self):
        return hash(str(astuple(self)))

    def dict(self):
        dictionary = super().dict()
        # Rename keys
        if dictionary.get('content_permissions') is not None:
            dictionary['contentPermissions'] = dictionary['content_permissions']
            del dictionary['content_permissions']
        return dictionary


@dataclass
class Group(ServerObject):
    """ A TableauServer Group object """
    name: str = None
    minimum_site_role: str = 'Viewer'
    domain_name: str = None
    import_domain_name: str = None
    import_site_role: str = None
    import_grant_license_mode: str = None
    user_count: int = None

    def __hash__(self):
        return hash(str(astuple(self)))

    def dict(self):
        dictionary = super().dict()
        # Rename keys
        if dictionary.get('minimum_site_role') is not None:
            dictionary['minimumSiteRole'] = dictionary['minimum_site_role']
            del dictionary['minimum_site_role']
        return dictionary


@dataclass
class Datasource(ServerObject):
    """ A TableauServer Datasource object """
    name: str = None
    created_at: datetime = None
    updated_at: datetime = None
    project_id: str = None
    project_name: str = None
    owner_id: str = None
    tags: list = None
    description: str = None
    content_url: str = None
    type: str = None
    encrypt_extracts: bool = None
    has_alert: bool = None
    has_extracts: bool = None
    is_certified: bool = None
    is_published: bool = None
    use_remote_query_agent: bool = None
    database_name: str = None
    server_name: str = None
    favorites_total: int = None
    connected_workbooks_count: int = None
    webpage_url: str = None
    certification_note: str = None

    def __hash__(self):
        return hash(str(astuple(self)))

    def publish_xml(self, conn_creds=None):
        xml_dict = {
            'tsRequest': {
                'datasource': {
                    '@name': self.name,
                    '@useRemoteQueryAgent': str(self.use_remote_query_agent).lower(),
                    '@description': self.description,
                    'project': {'@id': self.project_id}
                }
            }
        }
        if conn_creds:
            xml_dict['tsRequest']['datasource']['connectionCredentials'] = {
                '@name': conn_creds['username'],
                '@password': conn_creds['password'],
                '@embed': "true"
            }
        return xmltodict.unparse(xml_dict)


@dataclass
class Connection(ServerObject):
    """ A TableauServer Connection object """
    type: str = None
    datasource_id: str = None
    datasource_name: str = None
    workbook_id: str = None
    workbook_name: str = None
    user_name: str = None
    server_address: str = None
    server_port: int = None
    embed_password: bool = None
    password: str = None
    query_tagging_enabled: bool = None

    def __hash__(self):
        return hash(str(astuple(self)))

    def dict(self):
        dictionary = super().dict()

        del dictionary['type']
        if 'datasource_id' in dictionary:
            del dictionary['datasource_id']
        if 'datasource_name' in dictionary:
            del dictionary['datasource_name']
        return dictionary


@dataclass
class View(ServerObject):
    """ A TableauServer View object """
    name: str = None
    created_at: datetime = None
    updated_at: datetime = None
    workbook_id: str = None
    owner_id: str = None
    project_id: str = None
    tags: list = None
    location_id: str = None
    location_type: str = None
    content_url: str = None
    view_url_name: str = None
    sheet_type: str = None
    usage_total_view_count: int = None

    def __hash__(self):
        return hash(str(astuple(self)))


@dataclass
class Workbook(ServerObject):
    """ A TableauServer Workbook object """
    name: str = None
    created_at: datetime = None
    updated_at: datetime = None
    project_id: str = None
    project_name: str = None
    owner_id: str = None
    owner_name: str = None
    location_type: str = None
    location_name: str = None
    tags: list = None
    default_view_id: str = None
    description: str = None
    content_url: str = None
    webpage_url: str = None
    data_acceleration_config_acceleration_enabled: bool = None
    show_tabs: bool = None
    encrypt_extracts: bool = None
    has_extracts: bool = None
    views: list = None
    location_id: str = None
    size: int = 0
    sheet_count: int = None

    def __hash__(self):
        return hash(str(astuple(self)))

    def publish_xml(self, conns=None):
        xml_dict = {
            'tsRequest': {
                'workbook': {
                    '@name': self.name,
                    '@showTabs': str(self.show_tabs).lower(),
                    '@thumbnailsUserId': self.owner_id,
                    'project': {'@id': self.project_id}
                }
            }
        }
        if conns:
            xml_dict['tsRequest']['workbook'].setdefault('connections', list())
            for conn in conns:
                xml_dict['tsRequest']['workbook']['connections'].append({
                    'connection': {
                        '@serverAddress': conn['address'],
                        '@serverPort': conn['port'],
                        'connectionCredentials': {
                            '@name': conn['username'],
                            '@password': conn['password'],
                            '@embed': "true"
                        }
                    }
                })
        return xmltodict.unparse(xml_dict)


@dataclass
class Subscription(ServerObject):
    """ A TableauServer Subscription object """
    content_id: str = None
    content_type: str = None
    schedule_id: str = None
    schedule_name: str = None
    user_id: str = None
    user_name: str = None
    subject: str = None
    message: str = None
    page_orientation: str = None
    page_size_option: str = None
    content_send_if_view_empty: bool = None
    attach_image: bool = None
    attach_pdf: bool = None
    suspended: bool = None

    def __hash__(self):
        return hash(str(astuple(self)))


@dataclass
class Schedule(ServerObject):
    """ A TableauServer Schedule object """
    name: str = None
    state: str = None
    created_at: datetime = None
    updated_at: datetime = None
    next_run_at: datetime = None
    type: str = None
    frequency: str = None
    priority: int = None

    def __hash__(self):
        return hash(str(astuple(self)))


@dataclass
class Table(ServerObject):
    """ A TableauServer Table object """
    name: str = None
    site_id: str = None
    schema: str = None
    tags: list = None
    is_embedded: bool = None
    is_certified: bool = None

    def __hash__(self):
        return hash(str(astuple(self)))


@dataclass
class Database(ServerObject):
    """ A TableauServer Database object """
    name: str = None
    site_id: str = None
    connection_type: str = None
    type: str = None
    host_name: str = None
    content_permissions: str = None
    provider: str = None
    mime_type: str = None
    file_id: str = None
    request_url: str = None
    file_extension: str = None
    file_path: str = None
    tags: list = None
    is_embedded: bool = None
    is_certified: bool = None
    port: int = None

    def __hash__(self):
        return hash(str(astuple(self)))


@dataclass
class Job(ServerObject):
    """ A TableauServer Job object """
    created_at: datetime = None
    mode: str = None
    type: str = None
    updated_at: datetime = None
    completed_at: datetime = None
    finish_code: int = None
    extract_refresh_job_datasource_id: str = None
    extract_refresh_job_datasource_name: str = None
    id: str = None
    extract_refresh_job_notes_datasource_id: str = None
    extract_refresh_job_notes_datasource_name: str = None
    extract_refresh_job_notes_workbook_id: str = None
    extract_refresh_job_notes_workbook_name: str = None

    def __hash__(self):
        return hash(str(astuple(self)))


if __name__ == '__main__':
    u1 = User(id='1', name='Bob', full_name='Bob Johnson')
    u2 = User(id='1', name='Bob')
    u3 = User(id='2', name='Tom')
    for u in {u1, u2, u3}:
        print(u.dict())



import logging
import requests
import cgi
import os
import json
from urllib3.fields import RequestField
from urllib3.filepost import encode_multipart_formdata

import tableau_utilities.tableau_server.tableau_server_objects as tso
from tableau_utilities.general.funcs import flatten_dict


class TableauConnectionError(Exception):
    """ An Exception in the TableauServer connection """
    pass


class TableauServer:
    """ Connects and interacts with Tableau Online/Server, via the REST API. """

    def __init__(self, host, site, user=None, password=None, personal_access_token_secret=None, personal_access_token_name=None, api_version=None):
        """ To sign in to Tableau a user needs either a username & password or token secret & token name

        Args:
            host (str): Tableau server address
            user (str): The username to sign in to Tableau Online with
            password (str): The password to sign in to Tableau Online with
            personal_access_token_name (str): The name of the personal access token used
            personal_access_token_secret (str): The secret of the personal access token used
            site (str): The Tableau Online site id
            api_version (float): The Tableau REST API version
        """
        self.user = user
        self.__pw = password
        self.personal_access_token_secret = personal_access_token_secret
        self.personal_access_token_name = personal_access_token_name
        self.host = host
        self.site = site
        self.api = api_version or 3.18
        # Set by class
        self._auth_token = None
        self.url = None
        # Create a session on initialization
        self.session = requests.session()
        self.session.headers.update({'accept': 'application/json', 'content-type': 'application/json'})
        # Sign in on initialization
        self.__sign_in()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Sign out upon class deconstruction
        print('Signing out of Tableau Server')
        self.sign_out()

    def __sign_in(self):
        """
            To sign in to Tableau a user needs either a username & password or token secret & token name
            Signs in to the server with credentials from the specified connection.
            Sets the auth_token, site_id, and url common prefix

        """

        url = f"{self.host}/api/{self.api}/auth/signin"

        if self.personal_access_token_secret and self.personal_access_token_name:
            body = {"credentials": {"personalAccessTokenSecret": self.personal_access_token_secret,
                                    "personalAccessTokenName": self.personal_access_token_name,
                                    "site": {"contentUrl": self.site}}}
        elif self.user and self.__pw:
            body = {"credentials": {"name": self.user, "password": self.__pw, "site": {"contentUrl": self.site}}}
        else:
            raise TableauConnectionError(
                'Please provide either user and password, or token_secret and token_name'
            )

        res = self.post(url, json=body).get('credentials', {})
        # Set auth token and site ID attributes on sign in
        self.session.headers.update({'x-tableau-auth': res.get('token')})
        self.site = res.get('site', {}).get('id')
        self.url = f"{self.host}/api/{self.api}/sites/{self.site}"

    def sign_out(self):
        """ Destroys the active session and invalidates authentication token. """
        self.post(url=f"{self.host}/api/{self.api}/auth/signout")
        self.session.close()

    @staticmethod
    def __validate_response(response):
        """ Validates the response received from an API call

        Args:
            response (requests.Response): A requests Response object
            raise_for_status (bool): True to raise an error on a bad response

        Returns: The response content as a JSON dict
        """
        info = response.json()
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            error = info.get('error', {})
            raise TableauConnectionError(
                f'\nError: {error.get("code")}: {error.get("summary")} - {error.get("detail")}\n{err}'
            ) from err
        return info

    def get(self, url, headers=None, **params):
        """ GET request for the Tableau REST API

        Args:
            url (str): URL endpoint for GET call
            headers (dict): GET call header

        Returns: The response content as a JSON dict
        """

        res = self.session.get(url, headers=headers, **params)
        return self.__validate_response(res)

    def delete(self, url, headers=None, **params):
        """ DELETE request for the Tableau REST API

        Args:
            url (str): URL endpoint for DELETE call
            headers (dict): DELETE call header

        Returns: The response content as a JSON dict
        """
        res = self.session.delete(url, headers=headers, **params)
        return self.__validate_response(res)

    def post(self, url, json=None, headers=None, **params):
        """ POST request for the Tableau REST API

        Args:
            url (str): URL endpoint for POST call
            json (dict): The POST call JSON payload
            headers (dict): POST call header

        Returns: The response content as a JSON dict
        """
        res = self.session.post(url, json=json, headers=headers, **params)
        return self.__validate_response(res)

    def put(self, url, json=None, headers=None, **params):
        """ PUT request for the Tableau REST API

        Args:
            url (str): URL endpoint for PUT call
            json (dict): The PUT call JSON payload
            headers (dict): PUT call header

        Returns: The response content as a JSON dict
        """
        res = self.session.put(url, json=json, headers=headers, **params)
        return self.__validate_response(res)

    @staticmethod
    def __transform_tableau_object(object_dict):
        """ Transforms the object dict from a Tableau REST API call
        Args:
            object_dict (dict): The object dict from a Tableau REST API call
        """
        if object_dict.get('tags'):
            object_dict['tags'] = [t['label'] for t in object_dict['tags']['tag']]
        update = dict()
        flatten_dict(object_dict, update)
        object_dict.clear()
        object_dict.update(update)

    def __get_objects_pager(self, url, obj, page_size=100):
        """ GET all objects in the site
        Args:
            url: The url of for the request, i.e /api/api-version/sites/site-id/groups
            obj: The name of the object being requested
            page_size: The size of the page (number of objects per page)
        Returns: A list of the objects
        """
        response = self.get(url)
        total_available = response.get('pagination', {}).get('totalAvailable', page_size)
        total_available = int(total_available)
        current = 0
        page = 0
        while current < total_available:
            current += page_size
            page += 1
            if '?' in url:
                page_url = f'{url}&pageSize={page_size}&pageNumber={page}'
            else:
                page_url = f'{url}?pageSize={page_size}&pageNumber={page}'
            logging.info('GET %s --> %s/%s', page_url, current, total_available)
            res = self.get(page_url)
            for obj_dict in res[f'{obj}s'][obj]:
                self.__transform_tableau_object(obj_dict)
                yield obj_dict

    def get_datasource(self, datasource_id=None, datasource_name=None, datasource_project=None):
        """ Queries for a datasource in the site
            URI GET /api/api-version/sites/site-id/datasources/datasource_id

            (Optional) Can get the datasource either by ID, or by name & project.

        Args:
              datasource_id (str): The ID of the datasource
              datasource_name (str): The name of the datasource
              datasource_project (str): The name of the project the datasource is in

        Returns: A Datasource Tableau object
        """
        if datasource_id:
            d = self.get(f'{self.url}/datasources/{datasource_id}')
            d = d['datasource']
            self.__transform_tableau_object(d)
            return tso.Datasource(**d)
        elif datasource_name and datasource_project:
            for d in self.get_datasources():
                if d.name == datasource_name and d.project_name == datasource_project:
                    return d
            raise TableauConnectionError(
                f'Datasource not found:\n\tName    -> {datasource_name}\n\tProject -> {datasource_project}'
            )

        raise TableauConnectionError(
            'Please provide either the datasource_id, or both datasource_name and datasource_project'
        )

    def get_datasources(self):
        """ Queries for all datasources in the site
            URI GET /api/api-version/sites/site-id/datasources
        Returns: All datasources in the site
        """
        url = f"{self.url}/datasources?fields=_default_" \
              f",favoritesTotal" \
              f",databaseName" \
              f",connectedWorkbooksCount" \
              f",hasAlert" \
              f",hasExtracts" \
              f",isPublished" \
              f",serverName"

        for d in self.__get_objects_pager(url, 'datasource'):
            yield tso.Datasource(**d)

    def get_datasource_connections(self, datasource_id):
        """ Queries for all Connection Tableau objects in the datasource
            URI GET /api/api-version/sites/site-id/datasources/datasource_id/connections
        Args:
            datasource_id (str): The ID of the Tableau Datasource
        Returns: All Connection Tableau objects in the datasource
        """
        url = f'{self.url}/datasources/{datasource_id}/connections'
        for connection in self.__get_objects_pager(url, 'connection'):
            yield tso.Connection(**connection)

    def get_workbook_connections(self, workbook_id):
        """ Queries for all Connection Tableau objects in the workbook
            URI GET /api/api-version/sites/site-id/workbooks/workbook_id/connections
        Args:
            workbook_id (str): The ID of the Tableau Workbook
        Returns: All Connection Tableau objects in the workbook
        """
        url = f'{self.url}/workbooks/{workbook_id}/connections'
        for connection in self.__get_objects_pager(url, 'connection'):
            yield tso.Connection(**connection)

    def get_workbook(self, workbook_id=None, workbook_name=None, workbook_project=None):
        """ Queries for a workbook in the site
            URI GET /api/api-version/sites/site-id/workbooks/workbook_id
        Args:
            workbook_id (str): The ID of the Tableau Workbook
            workbook_name (str): The name of the workbook
            workbook_project (str): The name of the project the workbook is in
        Returns: A Workbooks Tableau object
        """
        if workbook_id:
            w = self.get(f'{self.url}/workbooks/{workbook_id}')
            w = w['workbook']
            self.__transform_tableau_object(w)
            return tso.Workbook(**w)
        elif workbook_name and workbook_project:
            for w in self.get_workbooks():
                if w.name == workbook_name and w.project_name == workbook_project:
                    return w
            raise TableauConnectionError(
                f'Datasource not found:\n\tName    -> {workbook_name}\n\tProject -> {workbook_project}'
            )

        raise TableauConnectionError(
            'Please provide either the workbook_id, or both workbook_name and workbook_project'
        )

    def get_workbooks(self):
        """ Queries for all workbooks in the site
            URI GET /api/api-version/sites/site-id/workbooks
        Returns: All workbooks in the site
        """
        for w in self.__get_objects_pager(f"{self.url}/workbooks", 'workbook'):
            yield tso.Workbook(**w)

    def get_view(self, view_id):
        """ Queries for a view in the site
            URI GET /api/api-version/sites/site-id/views/view_id
        Args:
            view_id (str): The ID of the Tableau View
        Returns: A View Tableau object
        """
        v = self.get(f'{self.url}/views/{view_id}')
        v = v['view']
        return tso.View(**v)

    def get_views(self):
        """ Queries for all views in the site
            URI GET /api/api-version/sites/site-id/views
        Returns: All workbooks in the site
        """
        url = f"{self.url}/views?fields=_default_,sheetType,usage"
        for v in self.__get_objects_pager(url, 'view'):
            yield tso.View(**v)

    def get_project(self, project_id):
        """ Queries for the project by project_id
            URI GET /api/api-version/sites/site-id/projects/project_id
        Args:
            project_id (str): The ID of the project in Tableau Online
        Returns: A Tableau Project object specified by ID
        """
        for p in self.get_projects():
            if p.id == project_id:
                return p

    def get_projects(self, top_level_only=True):
        """ Queries for all projects in the site
            URI GET /api/api-version/sites/site-id/projects
        Args:
            top_level_only (bool): True to only get top level projects
        Returns: All top level projects in the site
        """
        url = f"{self.url}/projects?fields=_default_" \
              f",topLevelProject" \
              f",writeable" \
              f",contentsCounts.projectCount" \
              f",contentsCounts.viewCount" \
              f",contentsCounts.datasourceCount" \
              f",contentsCounts.workbookCount"
        for p in self.__get_objects_pager(url, 'project'):
            project = tso.Project(**p)
            # Only get Top Level projects
            if top_level_only and not project.top_level_project:
                continue
            yield project

    def get_group(self, group_id):
        """ Queries for the group by user_id
            URI GET /api/api-version/sites/site-id/groups/group_id
        Args:
            group_id (str): The ID of the group in Tableau Online
        Returns: A Tableau Group object specified by ID
        """
        for group in self.get_groups():
            if group.id == group_id:
                return group

    def get_groups(self):
        """ Queries for all groups in the site
            URI GET /api/api-version/sites/site-id/groups
        Returns: All groups in the site
        """
        url = f"{self.url}/groups?fields=_default_,userCount,minimumSiteRole"
        for g in self.__get_objects_pager(url, 'group'):
            yield tso.Group(**g)

    def get_user(self, user_id):
        """ Queries for the user by user_id
            URI GET /api/api-version/sites/site-id/users/user_id
        Args:
            user_id (str): The ID of the user in Tableau Online
        Returns: A Tableau User object specified by ID
        """
        u = self.get(f"{self.url}/users/{user_id}")
        u = u['user']
        self.__transform_tableau_object(u)
        return tso.User(**u)

    def get_users(self):
        """ Queries for all users in the site
            URI GET /api/api-version/sites/site-id/users
        Returns: All users in the site
        """
        url = f"{self.url}/users?fields=_default_,fullName,email"
        for user in self.__get_objects_pager(url, 'user'):
            yield tso.User(**user)

    def get_user_groups(self):
        """ Queries for all groups and all user in those groups
            URI GET /api/api-version/sites/site-id/groups/group_id/users
        Returns: A list of all user/group combinations
        """
        for group in self.get_groups():
            for u in self.__get_objects_pager(f"{self.url}/groups/{group.id}/users", 'user'):
                yield group, tso.User(**u)

    def create_project(self, name, description='', content_permissions='LockedToProject'):
        """ Creates a project.

        Args:
            name (str): The name of the project
            description (str): The description of the project
            content_permissions (str): The content permissions, e.g. LockedToProject
        """
        self.post(
            f'{self.url}/projects',
            {
                'project': {
                    'name': name,
                    'description': description,
                    'contentPermissions': content_permissions
                }
            }
        )

    def create_group(self, name, minimum_site_role='Viewer'):
        """ Creates a group.

        Args:
            name (str): The name of the Group
            minimum_site_role (str): The minimum site role of the group, e.g. Viewer
        """
        self.post(
            f'{self.url}/groups',
            {
                'group': {
                    'name': name,
                    'minimumSiteRole': minimum_site_role
                }
            }
        )

    def __download_object(self, url, file_dir=None):
        """ Downloads a datasource from Tableau Online

        Args:
            url (str): The URL for the request
            file_dir (str): The file directory to write the file to

        Returns: The absolute path to the file
        """
        res = self.session.get(url, stream=True)
        try:
            res.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise TableauConnectionError(err) from err
        _, params = cgi.parse_header(res.headers["Content-Disposition"])
        file_name = os.path.basename(params["filename"])
        if file_dir:
            os.makedirs(file_dir, exist_ok=True)
            path = os.path.join(file_dir, file_name)
        else:
            path = file_name
        with open(path, "wb") as f:
            # Download in 1024 bytes (1kb) chunks
            for chunk in res.iter_content(1024):
                f.write(chunk)
        return os.path.abspath(path)

    def download_datasource(self, datasource_id, file_dir=None, include_extract=False):
        """ Downloads a datasource from Tableau Online

        Args:
            datasource_id (str):
            file_dir (str):
            include_extract (bool):
        """
        return self.__download_object(
            f'{self.url}/datasources/{datasource_id}/content?includeExtract={include_extract}',
            file_dir
        )

    def download_workbook(self, workbook_id, file_dir=None, include_extract=False):
        """ Downloads a workbook from Tableau Online

        Args:
            workbook_id (str):
            file_dir (str):
            include_extract (bool):
        """
        return self.__download_object(
            f'{self.url}/workbooks/{workbook_id}/content?includeExtract={include_extract}',
            file_dir
        )

    @staticmethod
    def __get_multipart_details(parts):
        """ Gets the body and content_type for a multipart/mixed request.

        Args:
            parts (list[tuple[str, str, str, str]]): The parts that make up the RequestField
                i.e. [(name, data, file_name, content_type)]

        Returns: Request body and content_type
        """
        part_list = list()
        for name, data, file_name, content_type in parts:
            r = RequestField(name=name, data=data, filename=file_name)
            r.make_multipart(content_type='application/octet-stream')
            part_list.append(r)
        post_body, content_type = encode_multipart_formdata(part_list)
        content_type = ''.join(('multipart/mixed',) + content_type.partition(';')[1:])
        return post_body, content_type

    def __get_datasource_for_publication(self, datasource_id, datasource_name, project_name):
        if datasource_id:
            return self.get_datasource(datasource_id)
        elif datasource_name and project_name:
            project = None
            for p in self.get_projects():
                if p.name == project_name:
                    project = p
                    break
            if not project:
                raise TableauConnectionError(f'Project does not exist: {project}')
            return tso.Datasource(
                name=datasource_name,
                project_id=project.id,
                project_name=project.name
            )
        else:
            raise TableauConnectionError('Specify datasource_id or datasource_name and project_name')

    def __get_workbook_for_publication(self, workbook_id, workbook_name, project_name):
        if workbook_id:
            return self.get_workbook(workbook_id)
        elif workbook_name and project_name:
            project = None
            for p in self.get_projects():
                if p.name == project_name:
                    project = p
                    break
            if not project:
                raise TableauConnectionError(f'Project does not exist: {project}')
            return tso.Workbook(
                name=workbook_name,
                project_id=project.id,
                project_name=project.name
            )
        else:
            raise TableauConnectionError('Specify datasource_id or datasource_name and project_name')

    def __upload_in_chunks(self, file_path, chunk_size_mb=5):
        """ Uplaods a file to Tableau, in chunks.
            - PUT /api/api-version/sites/site-id/fileUploads
            - PUT /api/api-version/sites/site-id/fileUploads/upload_session_id

        Args:
            file_path (str): The path to the file
            chunk_size_mb (int): The chunking size of increments to be uploaded

        Returns: An upload_session_id of the uploaded file
        """
        file_name = os.path.basename(file_path)
        # 1024 bytes in 1kb, 1000kb in 1mb
        chunk_size = 1024 * 1000 * chunk_size_mb
        # Initialize file upload session
        file = self.post(f'{self.url}/fileUploads')
        file = file['fileUpload']
        request_payload = RequestField(name='request_payload', data='')
        request_payload.make_multipart(content_type='text/xml')
        # Append file data in chunks
        with open(file_path, 'rb') as f:
            # Read in chunks
            for idx, chunk in enumerate(iter(lambda: f.read(chunk_size), b''), 1):
                post_body, content_type = self.__get_multipart_details([
                    ('request_payload', '', None, 'text/xml'),
                    ('tableau_file', chunk, file_name, 'application/octet-stream')
                ])
                self.put(
                    f'{self.url}/fileUploads/{file["uploadSessionId"]}',
                    data=post_body, headers={'Content-Type': content_type}
                )
        return file["uploadSessionId"]

    def publish_datasource(self, file_path, datasource_id=None, datasource_name=None, project_name=None, **kw):
        """ Publishes a datasource to Tableau Online.
            One of the following MUST be provided:
              - datasource_id: If the datasource already exists
              - datasource_name AND project name: If this is a new datasource

        Args:
            file_path (str): The path to the datasource file (.tds or .tdsx)
            datasource_id (str): The ID of the datasource in Tableau Online
            datasource_name (str): The name of the Datasource
            project_name (str): The name of the Project in Tableau Online

        Keyword Args:
            overwrite (bool): True to overwrite the datasource, if it exists
            as_job (bool): True to kick this off as an async job in Tableau Online
            append (bool): True to append the data to the datasource in Tableau Online
            connection (dict): A dict of connection credentials to embed in the datasource
                i.e. 'username' and 'password'

        Returns: A Datasource Tableau server object
        """
        overwrite = kw.pop('overwrite', True)
        as_job = kw.pop('as_job', False)
        append = kw.pop('append', False)
        connection = kw.pop('connection', None)
        file_name = os.path.basename(file_path)
        # 1024 bytes in 1kb, 1000kb in 1mb
        file_size_mb = os.path.getsize(file_path) / 1024 / 1000
        extension = file_path.split('.')[-1]
        datasource = self.__get_datasource_for_publication(datasource_id, datasource_name, project_name)
        ds_xml = datasource.publish_xml(connection)
        # Datasource must be 64mb or less to publish all at once
        maximum_megabytes = 64
        if file_size_mb > maximum_megabytes:
            upload_session_id = self.__upload_in_chunks(file_path, maximum_megabytes)
            publish_url = f'{self.url}/datasources?uploadSessionId={upload_session_id}' \
                          f'&datasourceType={extension}&overwrite={overwrite}&append={append}&asJob={as_job}'
            post_body, content_type = self.__get_multipart_details([
                ('request_payload', ds_xml, None, 'text/xml')
            ])
        else:
            publish_url = f'{self.url}/datasources?datasourceType={extension}' \
                          f'&overwrite={overwrite}&append={append}&asJob={as_job}'
            with open(file_path, 'rb') as f:
                post_body, content_type = self.__get_multipart_details([
                    ('request_payload', ds_xml, None, 'text/xml'),
                    ('tableau_datasource', f.read(), file_name, 'application/octet-stream')
                ])

        # Finally, publish the file uploaded
        content = self.post(publish_url, data=post_body, headers={'Content-Type': content_type})
        self.__transform_tableau_object(content['datasource'])
        return tso.Datasource(**content['datasource'])

    def publish_workbook(self, file_path, workbook_id=None, workbook_name=None, project_name=None, **kw):
        """ Publishes a workbook to Tableau Online.
            One of the following MUST be provided:
              - workbook_id: If the workbook already exists
              - workbook_name AND project name: If this is a new workbook

        Args:
            file_path (str): The path to the Workbook file (.twb or .twbx)
            workbook_id (str): The ID of the Workbook in Tableau Online
            workbook_name (str): The name of the Workbook
            project_name (str): The name of the Project in Tableau Online

        Keyword Args:
            overwrite (bool): True to overwrite the datasource, if it exists
            as_job (bool): True to kick this off as an async job in Tableau Online
            skip_connection_check (bool): True for Tableau server to not check if,
                a non-published connection, of a workbook is reachable
            connections (list[dict]): A list of connections
                i.e. [{address, port, username, password}]

        Returns: A Workbook Tableau server object
        """
        overwrite = kw.pop('overwrite', True)
        as_job = kw.pop('as_job', False)
        skip_connection_check = kw.pop('skip_connection_check', False)
        connections = kw.pop('connections', None)
        file_name = os.path.basename(file_path)
        # 1024 bytes in 1kb, 1000kb in 1mb
        file_size_mb = os.path.getsize(file_path) / 1024 / 1000
        extension = file_path.split('.')[-1]
        workbook = self.__get_workbook_for_publication(workbook_id, workbook_name, project_name)
        wb_xml = workbook.publish_xml(connections)
        # Datasource must be 64mb or less to publish all at once
        maximum_megabytes = 64
        if file_size_mb > maximum_megabytes:
            upload_session_id = self.__upload_in_chunks(file_path, maximum_megabytes)
            publish_url = f'{self.url}/workbooks?uploadSessionId={upload_session_id}' \
                          f'&workbookType={extension}' \
                          f'&skipConnectionCheck={skip_connection_check}' \
                          f'&overwrite={overwrite}&asJob={as_job}'
            post_body, content_type = self.__get_multipart_details([
                ('request_payload', wb_xml, None, 'text/xml')
            ])
        else:
            publish_url = f'{self.url}/workbooks?workbookType={extension}' \
                          f'&overwrite={overwrite}' \
                          f'&skipConnectionCheck={skip_connection_check}' \
                          f'&asJob={as_job}'
            with open(file_path, 'rb') as f:
                post_body, content_type = self.__get_multipart_details([
                    ('request_payload', wb_xml, None, 'text/xml'),
                    ('tableau_workbook', f.read(), file_name, 'application/octet-stream')
                ])

        # Finally, publish the file uploaded
        content = self.post(publish_url, data=post_body, headers={'Content-Type': content_type})
        self.__transform_tableau_object(content['workbook'])
        return tso.Workbook(**content['workbook'])

    def refresh_datasource(self, datasource_id):
        content = self.post(f'{self.url}/datasources/{datasource_id}/refresh', json={})
        self.__transform_tableau_object(content['job'])
        return tso.Job(**content['job'])

    def refresh_workbook(self, workbook_id):
        content = self.post(f'{self.url}/workbooks/{workbook_id}/refresh', json={})
        self.__transform_tableau_object(content['job'])
        return tso.Job(**content['job'])

    def update_datasource_connection(self, datasource_id, connection: tso.Connection):
        conn_dict = connection.dict()
        if 'user_name' in conn_dict:
            conn_dict.setdefault('userName', conn_dict.pop('user_name'))
        if 'embed_password' in conn_dict:
            conn_dict.setdefault('embedPassword', conn_dict.pop('embed_password'))
        if 'password' in conn_dict:
            conn_dict.setdefault('password', conn_dict.pop('password'))
        if 'server_address' in conn_dict:
            conn_dict.setdefault('serverAddress', conn_dict.pop('server_address'))
        if 'query_tagging_enabled' in conn_dict:
            conn_dict.setdefault('queryTaggingEnabled', conn_dict.pop('query_tagging_enabled'))

        content = self.put(
            f'{self.url}/datasources/{datasource_id}/connections/{connection.id}',
            json={'connection': conn_dict}
        )

        self.__transform_tableau_object(content['connection'])
        return tso.Connection(**content['connection'])

    def embed_datasource_credentials(self, datasource_id, credentials, connection_type):
        """ Embed the given credentials for all connections of a datasource of the given connection type.
            Only embeds Username and Password credentials.

        Args:
            datasource_id (str): The ID of the datasource
            connection_type (str): Type of connection you want to embed creds for, i.e. snowflake
            credentials (dict): The credentials dict to embed
                i.e. {'username': 'user', 'password': 'password'}
        """
        for cred in ['username', 'password']:
            if not credentials.get(cred):
                raise TableauConnectionError(f'Missing required credential: {cred}')

        for c in self.get_datasource_connections(datasource_id):
            if c.type.lower() == connection_type.lower():
                c.user_name = credentials['username']
                c.password = credentials['password']
                c.embed_password = True
                response = self.update_datasource_connection(datasource_id, c)
                return response
